# SpringBoot RestAPI + Validator 项目实验

## 运行效果

~~~ bash
# 测试 方法1验证器，不通过
curl -XGET localhost:8080/user/0
# 结果 {"msg":"handleUsage1.userID: userID 必须大于等于1"}


# 测试 方法1验证器，通过
curl -XGET localhost:8080/user/1
# 结果 {"msg":"welcome back, 1"}


# 测试 方法2验证器，不通过
curl \
-X POST \
-H 'content-type: application/json' \
--url 'localhost:8080/person' \
-d '{ "name": "" }'
# 结果 {"msg":"名称不能为空"}


# 测试 方法2验证器，不通过
curl \
-X POST \
-H 'content-type: application/json' \
--url 'localhost:8080/person' \
-d '{ "name": "luo" }'
# 结果 {"msg":"昵称不能为空"}


# 测试 方法2验证器，通过
curl \
-X POST \
-H 'content-type: application/json' \
--url 'localhost:8080/person' \
-d '{ "name": "luo", "nickname": "sam" }'
# 结果 {"name":"luo","nickname":"sam"}
~~~

## 步骤

### 1. [快速搭建 SpringBoot RestAPI 项目](./INIT_RESTAPI.md)

以下步骤都基于第一步搭建的项目操作。

### 2. [`app/build.gradle`](./app/build.gradle) 再次增加依赖

- SpringBoot starter validator 2.5.4

- lombok 1.18.20

~~~ gradle
dependencies {

  // ...

  // [增加] SpringBoot Starter Validator 模块整合包
  implementation 'org.springframework.boot:spring-boot-starter-validation:2.5.4'

  // [增加] lombok 辅助开发工具包 & 辅助编译工具包
  implementation 'org.projectlombok:lombok:1.18.20'
  annotationProcessor 'org.projectlombok:lombok:1.18.20'

}
~~~


### 3. 验证器方法1

- 控制器，新建 [`Usage1Controller.java`](./app/src/main/java/org/example/Usage1Controller.java)

~~~ java
@RestController
@Validated // 添加 @Validated
public class Usage1Controller {

  // @Min(1) : 要求 userID 最小为 1
  @GetMapping("/user/{id}")
  public Map<String, String> handleUsage1(
    @PathVariable(name = "id") 
    @Min(value = 1, message = "userID 必须大于等于1") 
    long userID
  ) {
    return new HashMap<>(){{
      put("msg", "welcome back, " + userID);
    }};
  }

}
~~~

- 错误处理，新建 [`Usage1ExceptionHandler.java`](./app/src/main/java/org/example/Usage1ExceptionHandler.java)

~~~ java
// @RestControllerAdvice 表示错误处理类
@RestControllerAdvice
public class Usage1ExceptionHandler {

  // 处理 ConstraintViolationException
  @ExceptionHandler(ConstraintViolationException.class)
  // 遇到这个异常时，响应码为 400
  @ResponseStatus(HttpStatus.BAD_REQUEST)
  public Map<String, String> handleUsage1Exception(ConstraintViolationException e) {
    // 获取异常信息然后输出 { "msg": "错误信息" }
    return new HashMap<>(){{
      put("msg", e.getMessage());
    }};
  }

}
~~~

### 4. 验证器方法2

- 控制器，新建 [`Usage2Controller.java`](./app/src/main/java/org/example/Usage2Controller.java)

~~~ java
@RestController
public class Usage2Controller {

  @PostMapping("/person")
  public Person person (@Valid @RequestBody Person person) {
    return person; // 原样返回 json
  }

  // POJO class (纯Java对象类)
  @lombok.Getter
  @lombok.Setter
  public static class Person {
    @NotEmpty // 要求名字不能空
    private String name;
    @NotEmpty // 要求昵称不能空
    private String nickname;
  }
}
~~~

- 错误处理，新建 [`Usage2ExceptionHandler.java`](./app/src/main/java/org/example/Usage2ExceptionHandler.java)

~~~ java
// @RestControllerAdvice 表示错误处理类
@RestControllerAdvice
public class Usage2ExceptionHandler {

  // 处理 MethodArgumentNotValidException
  @ExceptionHandler(MethodArgumentNotValidException.class)
  @ResponseStatus(HttpStatus.BAD_REQUEST)
  public Map<String, String> handleUsage2Exception(MethodArgumentNotValidException e) {
    return new HashMap<>(){{
      // Exception 有很多方法，这里只展示错误信息
      put("msg", e.getBindingResult().getFieldError().getDefaultMessage());
    }};
  }
}
~~~

### 5. 启动应用，测试结果

~~~ bash
./gradlew bootRun
~~~

### 6. 打包成独立运行的 Jar

~~~ bash
./gradlew bootJar
~~~
